#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include "guiutils.h"
#include "statictip.h"
#include "win.h"    
#include "wincb.h"
#include "winopcb.h"
#include "windpy.h"


gint WinDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);

gint WinOPIDEnterEventCB(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
);
gint WinOPIDLeaveEventCB(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
);

void WinToolBarItemCB(
	toolbar_item_struct *item, gint id, gpointer data
);
void WinToolBarItemEnterCB(
	toolbar_item_struct *item, gint id, gpointer data
);
void WinToolBarItemLeaveCB(
	toolbar_item_struct *item, gint id, gpointer data
);

gint WinManageTOCB(gpointer data);

gint WinEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
gint WinDpyEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
gint WinDSPViewEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);

gint WinScaleEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
void WinPositionValueChangedCB(GtkAdjustment *adj, gpointer data);
void WinVolumeValueChangedCB(GtkAdjustment *adj, gpointer data);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Win toplevel GtkWindow "delete_event" signal callback.
 */
gint WinDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	win_struct *win = WIN(data);
	if(win == NULL)
	    return(FALSE);

	WinCloseCB(widget, data);

	return(TRUE);
}


/*
 *	Win OPID "enter_notify_event" signal callback.
 */
gint WinOPIDEnterEventCB(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
)
{
	win_opid_struct *opid = WIN_OPID(data);
	if(opid == NULL)
	    return(FALSE);

	WinStatusMessage(opid->win, opid->tooltip, FALSE);
	return(TRUE);
}

/*
 *	Win OPID "leave_notify_event" signal callback.
 */
gint WinOPIDLeaveEventCB(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
)
{
	win_opid_struct *opid = WIN_OPID(data);
	if(opid == NULL)
	    return(FALSE);

	WinStatusMessage(opid->win, NULL, FALSE);
	return(TRUE);
}


/*
 *	Win Tool Bar Item callback.
 */
void WinToolBarItemCB(
	toolbar_item_struct *item, gint id, gpointer data
)
{
	win_opid_struct *opid = WIN_OPID(data);
	if(opid == NULL)
	    return;

	if(opid->func_cb != NULL)
	    opid->func_cb(
		NULL,		/* GtkWidget */
		opid->win	/* Win */
	    );
}

/*
 *	Win Tool Bar Item enter callback.
 */
void WinToolBarItemEnterCB(
	toolbar_item_struct *item, gint id, gpointer data
)
{
	WinOPIDEnterEventCB(NULL, NULL, data);
}

/*
 *	Win Tool Bar Item leave callback.
 */
void WinToolBarItemLeaveCB(
	toolbar_item_struct *item, gint id, gpointer data
)
{
	WinOPIDLeaveEventCB(NULL, NULL, data);
}


/*
 *	Manage timeout callback.
 */
gint WinManageTOCB(gpointer data)
{
	win_struct *win = WIN(data);
	if(win == NULL)
	    return(FALSE);

	/* Manage Y events and monitor sound object playing */
	WinManage(win);

	return(TRUE);
}

/*
 *	Win toplevel GtkWindow event signal callback.
 */
gint WinEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	gboolean press;
	guint keyval, state;
	GdkEventConfigure *configure;
	GdkEventExpose *expose;
	GdkEventFocus *focus;
	GdkEventKey *key;
	win_struct *win = WIN(data);
	if((widget == NULL) || (event == NULL) || (win == NULL))   
	    return(status);

	switch((gint)event->type)
	{
	  case GDK_CONFIGURE:
	    configure = (GdkEventConfigure *)event;
	    break;

	  case GDK_EXPOSE:
	    expose = (GdkEventExpose *)event;
	    break;

	  case GDK_FOCUS_CHANGE:
	    focus = (GdkEventFocus *)event;
	    break;

	  case GDK_KEY_PRESS:
	  case GDK_KEY_RELEASE:
	    key = (GdkEventKey *)event;
	    press = (event->type == GDK_KEY_PRESS) ? TRUE : FALSE;
	    keyval = key->keyval;
	    state = key->state;
#define DO_STOP_KEY_SIGNAL_EMIT {               \
 gtk_signal_emit_stop_by_name(                  \
  GTK_OBJECT(widget),                           \
  press ?                                       \
   "key_press_event" : "key_release_event"      \
 );                                             \
}
	    switch(keyval)
	    {
	      case GDK_space:
		if(press)
		{
		    WinPlay(win);
		}
		status = TRUE;
		break;

	      case GDK_Escape:
		if(press)
		{
		    if(WinIsPlaying(win))
		    {
			WinPause(win);
		    }
		    else
		    {
			WinUnmap(win);
		    }
		}
		status = TRUE;
		break;

	      case GDK_b:
	      case GDK_Left:
	      case GDK_KP_Left:
		if(press)
		{
		    GtkRange *range = (GtkRange *)win->position_scale;
		    GtkAdjustment *adj = (range != NULL) ?
			range->adjustment : NULL;
		    if(adj != NULL)
		    {
			gfloat v = adj->value - adj->step_increment;
			if(v > (adj->upper - adj->page_size))
			    v = adj->upper - adj->page_size;
			if(v < adj->lower)
			    v = adj->lower;
			gtk_adjustment_set_value(adj, v);
		    }
		}
		DO_STOP_KEY_SIGNAL_EMIT
		status = TRUE;
		break;

	      case GDK_f:
	      case GDK_Right:
	      case GDK_KP_Right:
		if(press)
		{
		    GtkRange *range = (GtkRange *)win->position_scale;
		    GtkAdjustment *adj = (range != NULL) ? 
			range->adjustment : NULL;
		    if(adj != NULL)
		    {
			gfloat v = adj->value + adj->step_increment;
			if(v > (adj->upper - adj->page_size))
			    v = adj->upper - adj->page_size;
			if(v < adj->lower)
			    v = adj->lower;
			gtk_adjustment_set_value(adj, v);
		    }
		}
		status = TRUE;
		break;

	      case GDK_B:
	      case GDK_Page_Up:
	      case GDK_KP_Page_Up:
		if(press)
		{
		    GtkRange *range = (GtkRange *)win->position_scale;
		    GtkAdjustment *adj = (range != NULL) ?
			range->adjustment : NULL;
		    if(adj != NULL)
		    {
			gfloat v = adj->value - adj->page_increment;
			if(v > (adj->upper - adj->page_size))
			    v = adj->upper - adj->page_size;
			if(v < adj->lower)
			    v = adj->lower;
			gtk_adjustment_set_value(adj, v);
		    }
		}
		status = TRUE;
		break;

	      case GDK_F:
	      case GDK_Page_Down:
	      case GDK_KP_Page_Down:
		if(press)
		{
		    GtkRange *range = (GtkRange *)win->position_scale;
		    GtkAdjustment *adj = (range != NULL) ?  
			range->adjustment : NULL;
		    if(adj != NULL)
		    {
			gfloat v = adj->value + adj->page_increment;
			if(v > (adj->upper - adj->page_size))
			    v = adj->upper - adj->page_size;
			if(v < adj->lower)
			    v = adj->lower;
			gtk_adjustment_set_value(adj, v);
		    }
		}    
		status = TRUE;
		break;

	      case GDK_Home:
	      case GDK_KP_Home:
		if(press)
		{
		    GtkRange *range = (GtkRange *)win->position_scale;
		    GtkAdjustment *adj = (range != NULL) ?
			range->adjustment : NULL;
		    if(adj != NULL)
		    {
			gfloat v = adj->lower;
			gtk_adjustment_set_value(adj, v);
		    }
		}
		status = TRUE;
		break;

	      case GDK_End:
	      case GDK_KP_End:
		if(press)
		{
		    GtkRange *range = (GtkRange *)win->position_scale;
		    GtkAdjustment *adj = (range != NULL) ?
			range->adjustment : NULL;
		    if(adj != NULL)
		    {
			gfloat v = adj->upper - adj->page_size;
			gtk_adjustment_set_value(adj, v);
		    }
		}
		status = TRUE;
		break;


	    }
#undef DO_STOP_KEY_SIGNAL_EMIT
	    break;  
	}

	return(status);
}

/*
 *	Win display GtkDrawingArea event signal callback.
 */
gint WinDpyEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	GdkEventConfigure *configure;
	GdkEventExpose *expose;
	GdkEventFocus *focus;
	GdkEventButton *button;
	win_struct *win = WIN(data);
	if((widget == NULL) || (event == NULL) || (win == NULL))
	    return(status);                                    

	switch((gint)event->type)
	{
	  case GDK_CONFIGURE:
	    configure = (GdkEventConfigure *)event;
	    GDK_PIXMAP_UNREF(win->display_pm)
	    win->display_pm = GDK_PIXMAP_NEW(
		configure->width, configure->height
	    );
	    status = TRUE;
	    break;

	  case GDK_EXPOSE:
	    expose = (GdkEventExpose *)event;
	    WinDpyDraw(win);
	    status = TRUE;
	    break;

	  case GDK_FOCUS_CHANGE:
	    focus = (GdkEventFocus *)event;
	    if(!GTK_WIDGET_HAS_FOCUS(widget) && focus->in)
		GTK_WIDGET_SET_FLAGS(widget, GTK_HAS_FOCUS);
	    else if(GTK_WIDGET_HAS_FOCUS(widget) && !focus->in)
		GTK_WIDGET_UNSET_FLAGS(widget, GTK_HAS_FOCUS);
	    WinDpyDraw(win);
	    status = TRUE;
	    break;

	  case GDK_BUTTON_PRESS:
	    button = (GdkEventButton *)event;
	    if(!GTK_WIDGET_HAS_FOCUS(widget) &&
	       GTK_WIDGET_CAN_FOCUS(widget)
	    )
		gtk_widget_grab_focus(widget);
	    switch(button->button)
	    {
	      case 1:
		break;   

	      case 2:
		break;   

	      case 3:
		if(win->lists_menu != NULL)
		{
		    GtkMenu *menu = GTK_MENU(win->lists_menu);
		    gtk_menu_popup(
			menu, NULL, NULL,
			NULL, NULL,
			button->button, button->time
		    );
		}
		break;
	    }
	    status = TRUE;
	    break;      

	}

	return(status);
}

/*
 *	Win DSPView event signal callback.
 */
gint WinDSPViewEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	GdkEventButton *button;
	win_struct *win = WIN(data);
	if((widget == NULL) || (event == NULL) || (win == NULL))
	    return(status);

	switch((gint)event->type)
	{
	  case GDK_BUTTON_PRESS:
	    button = (GdkEventButton *)event;
	    switch(button->button)
	    {
	      case 1:
		if(button->state & GDK_SHIFT_MASK)
		    WinDSPViewDisplayStylePrev(win);
		else
		    WinDSPViewDisplayStyleNext(win);
		status = TRUE;
		break;

	      case 3:
		if(win->dspview_style_menu != NULL)
		{
		    GtkMenu *menu = GTK_MENU(win->dspview_style_menu);
		    gtk_menu_popup(
			menu, NULL, NULL,
			NULL, NULL,
			button->button, button->time
		    );
		    status = TRUE;
		}
		break;
	    }
	    break;
	}

	return(status);
}


/*
 *	Sets or unsets the scale's status tip.
 */
static void WinScaleSetStatusTip(
	win_struct *win, GtkWidget *w, gboolean set
)
{
	/* Handle by widget */

	/* Position Scale */
	if(w == win->position_scale)
	{
	    GtkAdjustment *adj = GTK_RANGE_GET_ADJUSTMENT(w);

	    if(set && (adj != NULL) && (win->sample_rate > 0) &&
	       (win->sample_size > 0) && (win->channels > 0)
	    )
	    {
		const gint sample_size_bytes = MAX(win->sample_size / 8, 1);
		const gulong	nsamples = (gulong)(adj->upper - adj->lower),
				sample_pos = (gulong)(adj->value - adj->lower),
				sec = (gulong)(
				    sample_pos / win->sample_rate /
				    win->channels / sample_size_bytes
				);
		const gfloat coeff = (nsamples > 0l) ?
		    ((gfloat)sample_pos / (gfloat)nsamples) : 0.0f;
		gchar *s = g_strdup_printf(
		    "%.2i:%.2i:%.2i %i%% (%ld/%ld)",
		    (gint)(sec / 3600),
		    (gint)(sec / 60) % 60,
		    (gint)(sec) % 60,
		    (gint)(coeff * 100.0f),
		    sample_pos, nsamples
		);
		StaticTipSet(
		    w, s,
		    STATIC_TIP_ALIGN_POINTER_CENTER,
		    STATIC_TIP_ALIGN_WIDGET_MIN,
		    0, 0
		);
		g_free(s);
	    }
	    else
		StaticTipSet(w, NULL, 0, 0, 0, 0);
	}
	/* Front Left Volume Scale */
	else if(w == win->volume_fleft_scale)
	{
	    GtkAdjustment *adj = GTK_RANGE_GET_ADJUSTMENT(w);

	    if(set && (adj != NULL))
	    {
		gchar *s = g_strdup_printf(
		    "Front left volume %i%%",
		    (gint)((1.0f - adj->value) * 100.0f)
		);
		StaticTipSet(
		    w, s, 
		    STATIC_TIP_ALIGN_WIDGET_MAX,
		    STATIC_TIP_ALIGN_POINTER_CENTER,
		    0, 0
		);
		g_free(s);
	    }
	    else
		StaticTipSet(w, NULL, 0, 0, 0, 0);
	}
	/* Front Right Volume Scale */
	else if(w == win->volume_fright_scale)
	{
	    GtkAdjustment *adj = GTK_RANGE_GET_ADJUSTMENT(w);

	    if(set && (adj != NULL))
	    {
		gchar *s = g_strdup_printf(
		    "Front right volume %i%%",
		    (gint)((1.0f - adj->value) * 100.0f)
		);
		StaticTipSet(
		    w, s, 
		    STATIC_TIP_ALIGN_WIDGET_MAX,
		    STATIC_TIP_ALIGN_POINTER_CENTER,
		    0, 0
		);
		g_free(s);
	    }
	    else
		StaticTipSet(w, NULL, 0, 0, 0, 0);
	}
	/* Back Left Volume Scale */
	else if(w == win->volume_bleft_scale)
	{
	    GtkAdjustment *adj = GTK_RANGE_GET_ADJUSTMENT(w);

	    if(set && (adj != NULL))
	    {
		gchar *s = g_strdup_printf(
		    "Back left volume %i%%",
		    (gint)((1.0f - adj->value) * 100.0f)
		);
		StaticTipSet(
		    w, s,
		    STATIC_TIP_ALIGN_WIDGET_MAX,
		    STATIC_TIP_ALIGN_POINTER_CENTER,
		    0, 0
		);
		g_free(s);
	    }
	    else
		StaticTipSet(w, NULL, 0, 0, 0, 0);
	}
	/* Back Right Volume Scale */
	else if(w == win->volume_bright_scale)
	{
	    GtkAdjustment *adj = GTK_RANGE_GET_ADJUSTMENT(w);

	    if(set && (adj != NULL))
	    {
		gchar *s = g_strdup_printf(
		    "Back right volume %i%%",
		    (gint)((1.0f - adj->value) * 100.0f)
		);
		StaticTipSet(
		    w, s,
		    STATIC_TIP_ALIGN_WIDGET_MAX,
		    STATIC_TIP_ALIGN_POINTER_CENTER,
		    0, 0
		);
		g_free(s);
	    }
	    else
		StaticTipSet(w, NULL, 0, 0, 0, 0);
	}
}

/*
 *	Win GtkScale event signal callback.
 */
gint WinScaleEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	GdkEventButton *button;
	GdkEventMotion *motion;
	win_struct *win = WIN(data);
	if((widget == NULL) || (event == NULL) || (win == NULL))
	    return(status);

	switch((gint)event->type)
	{
	  case GDK_BUTTON_PRESS:
	    button = (GdkEventButton *)event;
	    WinScaleSetStatusTip(win, widget, TRUE);
	    status = TRUE;
	    break;

	  case GDK_BUTTON_RELEASE:
	    button = (GdkEventButton *)event;
	    WinScaleSetStatusTip(win, widget, FALSE);
	    status = TRUE;
	    break;

	  case GDK_MOTION_NOTIFY:
	    motion = (GdkEventMotion *)event;
	    if(GTK_RANGE(widget)->button != 0)
	    {
		WinScaleSetStatusTip(win, widget, TRUE);
		status = TRUE;
	    }
	    break;
	}

	return(status);
}

/*
 *	Win position GtkAdjustment "value_changed" signal callback.
 */
void WinPositionValueChangedCB(GtkAdjustment *adj, gpointer data)
{
	YID yid;
	win_struct *win = WIN(data);
	if((adj == NULL) || (win == NULL))
	    return;

	if(win->freeze_count > 0)
	    return;

	win->freeze_count++;

	yid = win->yid;

	/* Update sound object values if playing */
	if(yid != YIDNULL)
	{
	    YEventSoundPlay v;

	    v.flags = YPlayValuesFlagYID |
		      YPlayValuesFlagPosition;
	    v.yid = yid;
	    v.position = (YDataPosition)adj->value;

	    YSetPlaySoundObjectValues(win->con, yid, &v);
	}

	WinDpyQueueDraw(win);

	win->freeze_count--;
}

/*
 *	Win volume GtkAdjustment "value_changed" signal callback.
 */
void WinVolumeValueChangedCB(GtkAdjustment *adj, gpointer data)
{
	gint channels;
	gfloat *vol;
	YID yid;
	win_struct *win = WIN(data);
	if(win == NULL)
	    return;

	if(win->freeze_count > 0)
	    return;

	win->freeze_count++;

	channels = win->channels;
	yid = win->yid;

	/* Get new volume values for all channels */
	vol = WinGetVolume(win);
	if(vol == NULL)
	{
	    win->freeze_count--;
	    return;
	}

	/* Find out which volume scale this signal is for and update
	 * other volume scales to match volume due to number of channels
	 */
	/* Front Left */
	if(adj == GTK_RANGE_GET_ADJUSTMENT(win->volume_fleft_scale))
	{
	    const gint i = 0;
	    switch(channels)
	    {
	      case 4:
		break;
	      case 2:
		vol[2] = vol[i];
		break;
	      case 1:
		vol[1] = vol[i];
		vol[2] = vol[i];
		vol[3] = vol[i];
		break;
	    }
	}
	/* Front Right */
	else if(adj == GTK_RANGE_GET_ADJUSTMENT(win->volume_fright_scale))
	{
	    const gint i = 1;
	    switch(channels)
	    {
	      case 4:
		break;
	      case 2:
		vol[3] = vol[i];
		break;
	      case 1:
		vol[0] = vol[i];
		vol[2] = vol[i];
		vol[3] = vol[i];
		break;
	    }
	}
	/* Back Left */
	else if(adj == GTK_RANGE_GET_ADJUSTMENT(win->volume_bleft_scale))
	{
	    const gint i = 2;
	    switch(channels)
	    {
	      case 4:
		break;
	      case 2:
		vol[0] = vol[i];
		break;
	      case 1:
		vol[0] = vol[i];
		vol[1] = vol[i];
		vol[3] = vol[i];
		break;
	    }
	}
	/* Back Right */
	else if(adj == GTK_RANGE_GET_ADJUSTMENT(win->volume_bright_scale))
	{
	    const gint i = 3;
	    switch(channels)
	    {
	      case 4:
		break;
	      case 2: 
		vol[1] = vol[i];
		break;
	      case 1: 
		vol[0] = vol[i];
		vol[1] = vol[i];
		vol[2] = vol[i];
		break;
	    }
	}

	/* Update sound object values if playing */
	if(yid != YIDNULL)
	{
	    YEventSoundPlay v;

	    v.flags = YPlayValuesFlagYID |
		      YPlayValuesFlagVolume;
	    v.yid = yid;
	    v.left_volume = (Coefficient)vol[0];
	    v.right_volume = (Coefficient)vol[1];
	    v.back_left_volume = (Coefficient)vol[2];
	    v.back_right_volume = (Coefficient)vol[3];

	    YSetPlaySoundObjectValues(win->con, yid, &v);
	}

	/* Update volume scales */
	WinSetVolume(win, vol);

	WinDpyQueueDraw(win);

	win->freeze_count--;
}
