#include <string.h>
#include <gtk/gtk.h>
#include "../include/disk.h"
#include "guiutils.h"
#include "win.h"
#include "windpy.h"
#include "config.h"


static void WinDpyDrawStats(
	win_struct *win, GdkDrawable *drawable, GdkGC *gc,
	GtkStateType state, GtkStyle *style,
	GdkRectangle *rect
);
static void WinDpyDrawVolume(
	win_struct *win, GdkDrawable *drawable, GdkGC *gc,
	GtkStateType state, GtkStyle *style, 
	GdkRectangle *rect, const gfloat *volume
);

void WinDpyDraw(win_struct *win);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)	(((x) < 0) ? (-(x)) : (x))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *      Draws the stats.
 */
static void WinDpyDrawStats(
	win_struct *win, GdkDrawable *drawable, GdkGC *gc,
	GtkStateType state, GtkStyle *style,
	GdkRectangle *rect
)
{
	gint x, y, font_height;
	gchar s[256];
	GdkTextBounds b;

	gdk_gc_set_clip_rectangle(gc, rect);

#define DRAW_STRING(_font_,_c_,_s_)		{	\
 if((_font_) != NULL) {					\
  gdk_string_bounds((_font_), (_s_), &b);		\
  if((_c_) != NULL)					\
   gdk_gc_set_foreground(gc, (_c_));			\
  gdk_draw_string(					\
  drawable, (_font_), gc,				\
  x + b.lbearing, y + (_font_)->ascent,			\
  (_s_)							\
 );							\
 x += b.width;						\
 font_height = MAX(					\
  (_font_)->ascent + (_font_)->descent,			\
  font_height						\
 );							\
} }
#define DRAW_BITMAP(_bm_,_c_,_center_)		{	\
 if((_bm_) != NULL) {					\
  gint width, height;					\
  gdk_window_get_size((_bm_), &width, &height);		\
  gdk_gc_set_clip_mask(gc, (_bm_));			\
  if(_center_)						\
   gdk_gc_set_clip_origin(				\
    gc, x - (width / 2), y - (height / 2)		\
   );							\
  else							\
   gdk_gc_set_clip_origin(gc, x, y);			\
  if((_c_) != NULL)					\
   gdk_gc_set_foreground(gc, (_c_));			\
  if(_center_)                                          \
   gdk_draw_rectangle(					\
    drawable, gc, TRUE,					\
    x - (width / 2), y - (height / 2), width, height	\
   );							\
  else							\
   gdk_draw_rectangle(					\
    drawable, gc, TRUE,					\
    x, y, width, height					\
   );							\
  gdk_gc_set_clip_mask(gc, NULL);			\
  if(_center_)						\
   x += width / 2;					\
  else							\
   x += width;						\
} }

	x = rect->x;
	y = rect->y;
	font_height = 0;
	if(win->cur_path != NULL)
	{
	    GtkRange *range = (GtkRange *)win->position_scale;
	    GtkAdjustment *adj = (range != NULL) ? range->adjustment : NULL;
	    const gulong	sample_pos = (adj != NULL) ?
				    (gulong)(adj->value - adj->lower) : 0l,
				nsamples = (adj != NULL) ?
				    (gulong)(adj->upper - adj->lower) : 0l;
 	    const gint	sample_rate = win->sample_rate,
			sample_size = win->sample_size,
			sample_size_bytes = sample_size / 8,
			channels = win->channels;
#if 0
	    const gchar *name = strrchr(win->cur_path, DIR_DELIMINATOR);
	    if(name != NULL)
		name++;
	    else
		name = win->cur_path;
	    DRAW_STRING(win->font_bold, &win->color_fg, name)
	    x = rect->x;
	    y += font_height + 1;
	    font_height = 0;
#endif

	    /* Position */
	    if((adj != NULL) && (sample_rate > 0) &&
	       (sample_size_bytes > 0) && (channels > 0)
	    )
	    {
		/* Audio values are known, suggesting a DSP sound object
		 * is loaded, so draw stats for the DSP sound object
		 */
		const gfloat coeff = (nsamples > 0l) ?
		    (gfloat)sample_pos / (gfloat)nsamples : 0.0f;
		const gulong sec = (gulong)(
		    sample_pos / sample_rate / channels / sample_size_bytes
		);

		/* Draw play/pause/stop icon */
		if(win->yid != YIDNULL)
		    DRAW_BITMAP(win->play_bm, &win->color_fg, FALSE)
		else if(sample_pos > 0)
		    DRAW_BITMAP(win->pause_bm, &win->color_fg, FALSE)
		else
		    DRAW_BITMAP(win->stop_bm, &win->color_fg, FALSE)
		x += 10;
		y += 0;

		/* Draw position in time label */
		snprintf(
		    s, sizeof(s),
		    "%.2i:%.2i:%.2i",
		    (gint)(sec / 3600),
		    (gint)(sec / 60) % 60,
		    (gint)(sec) % 60
		);
		DRAW_STRING(win->font_bold, &win->color_fg, s)
		x += 10;

		/* Draw position in percent label */
		snprintf(
		    s, sizeof(s),
		    "%i%%",
		    (gint)(coeff * 100.0f)
		);
		DRAW_STRING(win->font_bold, &win->color_fg, s)
		x = rect->x;
		y += font_height + 1;
		font_height = 0;
	    }
	    else
	    {
		/* Draw play/pause/stop icon */
                if(win->yid != YIDNULL)
                    DRAW_BITMAP(win->play_bm, &win->color_fg, FALSE)
                else if(sample_pos > 0)
                    DRAW_BITMAP(win->pause_bm, &win->color_fg, FALSE)
                else
                    DRAW_BITMAP(win->stop_bm, &win->color_fg, FALSE)
                x += 10;

                /* Draw MIDI icon */
		DRAW_BITMAP(win->midi_bm, &win->color_fg, FALSE)
                x += 10;
                y += 0; 
	    }

	    /* Sample rate & sample size */
	    if(sample_rate > 0)
	    {
		snprintf(
		    s, sizeof(s),
		    "%i Hz", sample_rate
		);
		DRAW_STRING(win->font, &win->color_fg, s)
		x += 10;
	    }
	    if(sample_size > 0)
	    {
		snprintf(
		    s, sizeof(s),
		    "%i bits", sample_size
		);
		DRAW_STRING(win->font, &win->color_fg, s)
		x = rect->x;
		y += font_height + 1;
		font_height = 0;
	    }

	    /* Channels */
	    switch(channels)
	    {
	      case 4:
		DRAW_STRING(win->font, &win->color_fg, "Quad") 
		break;
	      case 2:
		DRAW_STRING(win->font, &win->color_fg, "Stereo")
		break;
	      case 1:
		DRAW_STRING(win->font, &win->color_fg, "Mono")
		break;
	    }
	    x = rect->x;
	    y += font_height + 1;
	    font_height = 0;
	}
	else
	{
	    GdkFont *font = win->font_small;
	    const gchar *s = "(No Sound Object)";

	    if(font != NULL)
	    {
		GdkTextBounds b;

		gdk_string_bounds(font, s, &b);
		font_height = font->ascent + font->descent;

		gdk_gc_set_foreground(gc, &win->color_fg);
		gdk_draw_string(
		    drawable, font, gc,
		    rect->x +
			(rect->width / 2) - (b.width / 2) +
			b.lbearing,
		    rect->y +
			(rect->height / 2) - (font_height / 2) + 
			font->ascent,
		    s
		);
	    }
	}


	gdk_gc_set_clip_rectangle(gc, NULL);

#undef DRAW_STRING
#undef DRAW_BITMAP
}


/*
 *	Draws the volume ticks.
 */
static void WinDpyDrawVolume(
	win_struct *win, GdkDrawable *drawable, GdkGC *gc,
	GtkStateType state, GtkStyle *style,
	GdkRectangle *rect, const gfloat *volume
)
{
	GdkFont *font = win->font_small;
	const gint	nticks = WIN_NVOL_TICKS,
			tick_spacing = 2;
	gint	x, y,
		label_width = (font != NULL) ? (3 * 6) : 0,
		tick_width = MAX(
		    (rect->width - label_width -
			(tick_spacing * (nticks - 1))) /
			nticks,
		    1
		) + 0.5f,
		tick_height = 8;

	gdk_gc_set_clip_rectangle(gc, rect);

#define DRAW_VOL(_i_,_s_,_x_,_y_)	{		\
 gint n;						\
 GdkTextBounds b;					\
 const gfloat vol = volume[(_i_)];			\
 const gint end_tick = (vol * nticks) + 0.5f;		\
							\
 /* Draw label */					\
 if(font != NULL) {					\
  gdk_string_bounds(font, (_s_), &b);			\
  gdk_gc_set_foreground(gc, &win->color_fg);		\
  gdk_draw_string(					\
   drawable, font, gc,					\
   (_x_) + b.lbearing, (_y_) + font->ascent,		\
   (_s_)						\
  );							\
 }							\
 (_x_) += label_width;					\
							\
 /* Draw each tick for this volume */			\
 for(n = 0;						\
     n < nticks;					\
     n++, (_x_) += tick_width + tick_spacing		\
 ) {							\
  /* Set tick color */					\
  gdk_gc_set_foreground(				\
   gc,							\
   (n < end_tick) ? &win->color_vol_ticks[n] :		\
    &win->color_vol_ticks_dark[n]			\
  );							\
  /* Draw tick */					\
  gdk_draw_rectangle(					\
   drawable, gc, TRUE,					\
   (_x_), (_y_), tick_width, tick_height		\
  );							\
 }							\
}

	x = rect->x;
	y = rect->y + MAX(
	    (rect->height / 2) -
	    (((4 * tick_height) + (3 * tick_spacing)) / 2),
	    0
	);
	DRAW_VOL(2, "BL", x, y)

	x = rect->x;
	y += tick_height + tick_spacing;
	DRAW_VOL(0, "FL", x, y)

	x = rect->x;
	y += tick_height + tick_spacing;
	DRAW_VOL(1, "FR", x, y)

	x = rect->x;
	y += tick_height + tick_spacing;
	DRAW_VOL(3, "BR", x, y)

	gdk_gc_set_clip_rectangle(gc, NULL);

#undef DRAW_VOL
}


/*
 *	Redraws the Win's display.
 */
void WinDpyDraw(win_struct *win)
{
	gfloat volume[4];
	gint width, height;
	GdkRectangle rectangle, *rect = &rectangle;
	GdkWindow *window;
	GdkPixmap *pixmap;
	GdkDrawable *drawable;
	GdkGC *gc;
	GtkStateType state;
	GtkStyle *style;
	GtkWidget *w = (win != NULL) ? win->display_da : NULL;
	if(w == NULL)
	    return;

	window = w->window;
	pixmap = win->display_pm;
	gc = win->gc;
	state = GTK_WIDGET_STATE(w);
	style = gtk_widget_get_style(w);
	if(!GTK_WIDGET_VISIBLE(w) ||
	   (window == NULL) || (gc == NULL) || (style == NULL)
	)
	    return;

	drawable = (pixmap != NULL) ? pixmap : window;

	gdk_window_get_size(drawable, &width, &height);
	if((width <= 0) || (height <= 0))
	    return;

	/* Calculate volume; left, right, back left, and back right */
	volume[0] = 0.0f;
	volume[1] = 0.0f;
	volume[2] = 0.0f;
	volume[3] = 0.0f;
	if(win->yshm_buf != NULL)
	{
	    const guint8	*p = win->yshm_buf,
				*p_end = p + win->yshm_buf_len;

	    switch(win->channels)
	    {
	      case 1:
		while(p < p_end)
		{
		    volume[0] = MAX(
			volume[0],
			ABSOLUTE(((gfloat)*p - 128.0f) / 128.0f)
		    );
		    p++;
		}
		volume[1] = volume[2] = volume[3] = volume[0];
		break;

	      case 2:
		while(p < p_end)
		{
		    volume[0] = MAX(
			volume[0],
			ABSOLUTE(((gfloat)*p - 128.0f) / 128.0f)
		    );
		    p++;
		    if(p >= p_end)
			break;
		    volume[1] = MAX(
			volume[1],
			ABSOLUTE(((gfloat)*p - 128.0f) / 128.0f)
		    );
		    p++;
		}
		volume[2] = volume[0];
		volume[3] = volume[1];
		break;

	      case 4:
		while(p < p_end)
		{
		    volume[0] = MAX(
			volume[0],
			ABSOLUTE(((gfloat)*p - 128.0f) / 128.0f)
		    );
		    p++;
		    if(p >= p_end)
			break;
		    volume[1] = MAX(
			volume[1],
			ABSOLUTE(((gfloat)*p - 128.0f) / 128.0f)
		    );      
		    p++;
		    if(p >= p_end)
			break;
		    volume[2] = MAX(
			volume[2],
			ABSOLUTE(((gfloat)*p - 128.0f) / 128.0f)
		    );      
		    p++;
		    if(p >= p_end)
			break;
		    volume[3] = MAX(
			volume[3],
			ABSOLUTE(((gfloat)*p - 128.0f) / 128.0f)
		    );
		    p++;
		}
		break;
	    }
	}

	/* Draw base */
	gdk_gc_set_foreground(gc, &win->color_bg);
	gdk_draw_rectangle(
	    drawable, gc, TRUE,
	    0, 0, width, height
	);

	/* Draw stats */
	rect->width = width - 90 - (3 * 2);
	rect->height = height - (2 * 2);
	rect->x = 5;
	rect->y = 2;
	WinDpyDrawStats(win, drawable, gc, state, style, rect);

	/* Draw volume ticks */
	rect->width = 90;
	rect->height = height;
	rect->x = width - rect->width;
	rect->y = 0;
	WinDpyDrawVolume(win, drawable, gc, state, style, rect, volume);





	/* Draw focus rectangle if widget is in focus */
	if(GTK_WIDGET_HAS_FOCUS(w) && GTK_WIDGET_SENSITIVE(w))
	    gtk_draw_focus( 
		style, drawable,
		0, 0, width - 1, height - 1
	    );

	/* Send drawable to window if drawable is not the window */
	if(drawable != window)
	    gdk_draw_pixmap(
		window, style->fg_gc[state], drawable,
		0, 0, 0, 0, width, height
	    );
}

/*
 *	Qeueus a redraw of the Win's display.
 */
void WinDpyQueueDraw(win_struct *win)
{
	GtkWidget *w = (win != NULL) ? win->display_da : NULL;
	if(w != NULL)
	    gtk_widget_queue_draw(w);
}

